print("import: 'montreal_forced_aligner'")
import montreal_forced_aligner

print("import: 'montreal_forced_aligner.acoustic_modeling'")
import montreal_forced_aligner.acoustic_modeling

print("import: 'montreal_forced_aligner.alignment'")
import montreal_forced_aligner.alignment

print("import: 'montreal_forced_aligner.command_line'")
import montreal_forced_aligner.command_line

print("import: 'montreal_forced_aligner.corpus'")
import montreal_forced_aligner.corpus

print("import: 'montreal_forced_aligner.dictionary'")
import montreal_forced_aligner.dictionary

print("import: 'montreal_forced_aligner.g2p'")
import montreal_forced_aligner.g2p

print("import: 'montreal_forced_aligner.ivector'")
import montreal_forced_aligner.ivector

print("import: 'montreal_forced_aligner.language_modeling'")
import montreal_forced_aligner.language_modeling

print("import: 'montreal_forced_aligner.transcription'")
import montreal_forced_aligner.transcription

print("import: 'montreal_forced_aligner._version'")
import montreal_forced_aligner._version

print("import: 'montreal_forced_aligner.abc'")
import montreal_forced_aligner.abc

print("import: 'montreal_forced_aligner.config'")
import montreal_forced_aligner.config

print("import: 'montreal_forced_aligner.data'")
import montreal_forced_aligner.data

print("import: 'montreal_forced_aligner.exceptions'")
import montreal_forced_aligner.exceptions

print("import: 'montreal_forced_aligner.helper'")
import montreal_forced_aligner.helper

print("import: 'montreal_forced_aligner.models'")
import montreal_forced_aligner.models

print("import: 'montreal_forced_aligner.segmenter'")
import montreal_forced_aligner.segmenter

print("import: 'montreal_forced_aligner.speaker_classifier'")
import montreal_forced_aligner.speaker_classifier

print("import: 'montreal_forced_aligner.textgrid'")
import montreal_forced_aligner.textgrid

print("import: 'montreal_forced_aligner.utils'")
import montreal_forced_aligner.utils

print("import: 'montreal_forced_aligner.validator'")
import montreal_forced_aligner.validator

